#ifndef __FS_INTERFACE_HXX__
#define __FS_INTERFACE_HXX__

// depends on DWORDLONG and DWORD...

// disposition:
namespace FsInterface {

    typedef DWORDLONG UINT64;
    typedef DWORD UINT32;

    typedef UINT64 TIME64; // defined in units of 100ns since Jan 1, 1601 (AD)

    // disposition:
    const UINT32 DISP_CREATE_NEW        = 0x10000000;
    const UINT32 DISP_CREATE_ALWAYS     = 0x20000000;
    const UINT32 DISP_OPEN_EXISTING     = 0x30000000;
    const UINT32 DISP_OPEN_ALWAYS       = 0x40000000;
    const UINT32 DISP_TRUNCATE_EXISTING = 0x50000000;
    const UINT32 DISP_DIRECTORY         = 0x60000000;
    const UINT32 DISP_MASK              = 0x70000000;

    // access:
    const UINT32 ACCESS_READ         = 0x00010000;
    const UINT32 ACCESS_WRITE        = 0x00020000;
    const UINT32 ACCESS_MASK         = 0x00030000;

    // sharing:
    const UINT32 SHARE_READ          = 0x00100000;
    const UINT32 SHARE_WRITE         = 0x00200000;
    const UINT32 SHARE_MASK          = 0x00300000;

    // flags = dispositions | access | sharing
    const UINT32 FLAGS_MASK = DISP_MASK | ACCESS_MASK | SHARE_MASK;

    // attributes:
    const UINT32 ATTR_SYMLINK        = 0x00002000;
    const UINT32 ATTR_DIRECTORY      = 0x00000010;
    const UINT32 ATTR_READONLY       = 0x00000001;
    const UINT32 ATTR_HIDDEN         = 0x00000002;
    const UINT32 ATTR_SYSTEM         = 0x00000004;
    const UINT32 ATTR_ARCHIVE        = 0x00000020;
    const UINT32 ATTR_COMPRESSED     = 0x00000800;
    const UINT32 ATTR_OFFLINE        = 0x00001000;
    const UINT32 ATTR_MASK           = ATTR_SYMLINK | ATTR_DIRECTORY | ATTR_READONLY | ATTR_HIDDEN | ATTR_SYSTEM | ATTR_ARCHIVE | ATTR_COMPRESSED | ATTR_OFFLINE; // 0x00003837;

    const UINT64 INVALID_UINT64 = ((UINT64)(-1));
    const UINT32 INVALID_UINT32 = ((UINT32)(-1));
    const TIME64 INVALID_TIME64 = INVALID_UINT64;

    const size_t MAX_FS_NAME_LEN = 64; // somewhat arbitrary...

    struct fs_attr_t {
        char fs_name[MAX_FS_NAME_LEN];
        UINT64 total_bytes;
        UINT64 free_bytes;
    };

    // passing in INVALID_UINTxx for any of the values below makes it 
    // unspecified.

    // we do not use gid/uid because SMB does not deal...
    struct fattr_t {
        // sizes
        UINT64 file_size;
        UINT64 alloc_size;
        // times
        TIME64 create_time;
        TIME64 access_time;
        TIME64 mod_time;
        // mode/attr
        UINT32 attributes;
        // a la Win32, but wider so that we can pass in -1 safely
    };

    typedef UINT64 fhandle_t;

    // fhandle_t of 0 is a special file handle representing the root of that
    // file system

    const fhandle_t INVALID_FHANDLE_T = ((fhandle_t)(-1));

    const size_t MAX_NAME_LENGTH = 255;

    typedef struct {
        char name[MAX_NAME_LENGTH];
        struct fattr_t attribs;
        UINT32 cookie;
    } dirinfo_t;

    class FsDispatchTable {
    public:
        virtual ULONG AddRef() = 0;
        virtual ULONG Release() = 0;

        virtual const char* get_principal() = 0;
        virtual fhandle_t get_root() = 0;
        // returns immutable root handle, which can only be used by create(),
        // possibly with a NULL argument to signify an open of the root dir.

//         virtual DWORD create2(const char* path, UINT32 flags, fattr_t* attr, fhandle_t* handle) = 0;

        virtual DWORD create(fhandle_t dir, const char* name, UINT32 flags, fattr_t* attr, fhandle_t* handle) = 0;
        virtual DWORD set_attr(fhandle_t handle, fattr_t* attr) = 0;
        virtual DWORD lookup(fhandle_t dir, const char* name, fattr_t* attr) = 0;
        virtual DWORD get_attr(fhandle_t handle, fattr_t* attr) = 0;
        virtual DWORD close(fhandle_t handle) = 0;
        virtual DWORD write(fhandle_t handle, UINT64 offset, UINT64* count, void* buffer) = 0;
        virtual DWORD read(fhandle_t handle,  UINT64 offset, UINT64* count, void* buffer) = 0;
        virtual DWORD readlink(fhandle_t handle, int* size, char* path_buffer) = 0;
        virtual DWORD symlink(fhandle_t dir, const char* name, const char* path) = 0;
        // returns ERROR_NO_MORE_FILES when done...
        virtual DWORD read_dir(fhandle_t dir, UINT32 cookie, dirinfo_t* buffer, UINT32 size, UINT32 *entries_found) = 0;
        virtual DWORD statfs(fhandle_t handle, fs_attr_t* attr) = 0;
        virtual DWORD remove(fhandle_t dir, const char* name) = 0;
        virtual DWORD rename(fhandle_t fromdir, const char* fromname, fhandle_t todir, const char* toname) = 0;
        // we skip link
        virtual DWORD mkdir(fhandle_t dir, const char* name, fattr_t* attr) = 0;
        virtual DWORD rmdir(fhandle_t dir, const char* name) = 0;
//        virtual DWORD disconnect() = 0;
    };

    // to use readir, we open a dir with DISP_DIRECTORY...and remember to close
    // when done...

//    DWORD GetFileSystem(FileSystem** ppFs);
    class FileSystem {
    public:
        virtual ULONG AddRef() = 0;
        virtual ULONG Release() = 0;

        virtual DWORD connect(const char* principal, FsDispatchTable** ppDT) = 0;
    };

    typedef DWORD (WINAPI *FS_CREATE_PROC)(LPCSTR, LPCSTR, DWORD, FileSystem**);

    const char FS_CREATE_PROC_NAME[] = "FileSystemCreate";

    const DWORD FS_VERSION_0 = 0;
}

#endif /* __FS_INTERFACE_HXX__ */
